(* Mathematica package *)

BeginPackage["QuantumAlgebra`Options`"]



QAOptions::usage =
	"QAOptions[options] Invoke this to setup options for the application different from the default options. \n\n" <>
	"5 Options are defined: the 4 print options Verbosity, MaxLines, PrintStack, StackPattern and the DefinitionLevel. " <>
	"This function also resets the lines already printed to 0."

If[!ValueQ[MaxLines],
	MaxLines::usage = 
		"MaxLines -> max. The MaxLines option. Infinity by default. Restricts the number of printed diagnostic lines."
]

If[!ValueQ[PrintStack],
	PrintStack::usage = 
		"PrintStack -> opt. The PrintStack option: None, Full, Recursive."
]

If[!ValueQ[StackPattern],
	StackPattern::usage = 
		"StackPattern -> pattern. The StackPattern option: None by default."
]

If[!ValueQ[Verbosity],
	Verbosity::usage = 
		"Verbosity -> opt. The Verbosity option: 0 by default. Defines the Print Levels to print out."
]

If[!ValueQ[DefinitionLevel],
	DefinitionLevel::usage = 
		"DefinitionLevel -> lvl. The DefinitionLevel option: 3 by default.\n\n" <>
		"There are several Definitions of certain predicates integrated in this application. " <>
		"You can choose one by calling QAOptions[DefinitionLevel->lvl]. 3 will be the last " <>
		"and most sophisticated one."
]

If[!ValueQ[Recursive],
	Recursive::usage = 
		"Recursive. A value for the PrintStack option."
]

If[!ValueQ[PrtLevel],
	PrtLevel::usage = 
		"PrtLevel -> level. The Level to associate with a Print, 1 per default."
]

If[!ValueQ[Indent],
	Indent::usage = 
		"Indent -> value. The Indentation wo be used for Print, 1 per default (no indentation)."
]

QAPrint::usage =
	"QAPrint[ex__, options] is print function controlled by options."

QATiming::usage =
	"QATiming[ex] measures the calculation of the expression ex and signs it with a time stamp. \n\n" <>
	"It is compatible with Timing except that an additional time stamp will be produced. This can mark " <>
	"the calculation in a notebook so one knows the last time of execution."

QASniff::usage = 
	"QASniff[ex, desc] prints out an expression at a certain stage of an evaluation chain and returns the expression unchanged. \n\n" <>
	"ex is the expression in question and desc a description which defaults to 'Sniffing...' to mark the print out."


Begin["`Private`"]

(* ::Section:: *)
(* Region Print and Logging Support *)

Options[QAOptions] = 
{
	Verbosity -> 0,
	MaxLines -> Infinity,
	PrintStack -> None,
	StackPattern -> None,
	
	DefinitionLevel -> 3
}

Options[QAPrint] = 
{
	Indent -> 1,
	PrtLevel -> 1
}


(* Used to pre-set the Options for QAPrint or other purposes *)
QAOptions[OptionsPattern[]] :=
	Module[{},
		linesAlreadyPrinted = 0;
		SetOptions[
			QAOptions, 
			{ 
				Verbosity -> OptionValue[Verbosity], 
				MaxLines -> OptionValue[MaxLines],
				PrintStack -> OptionValue[PrintStack], 
				StackPattern -> OptionValue[StackPattern],
	
				DefinitionLevel -> OptionValue[DefinitionLevel] 
			}]
	]


(* Prints diagnostic expressions depending on set Options *)
SetAttributes[QAPrint, { HoldAll }]

QAPrint[expr___, OptionsPattern[]] := 
	Module[
		{
			prtlvl = OptionValue[QAPrint, PrtLevel],
			indent = StringJoin @@ Table["  ", OptionValue[QAPrint, Indent] - 1],
			
			maxlines = OptionValue[QAOptions, MaxLines],
			ps = OptionValue[QAOptions, PrintStack],
			sp = OptionValue[QAOptions, StackPattern],
			verbosity = OptionValue[QAOptions, Verbosity],
			tr
		},
		tr = 
			Which[
				ps =!= None && sp === None, 
				Stack[], 
				ps =!= None, 
				Stack[sp]
			];
		If[linesAlreadyPrinted < maxlines && prtlvl <= verbosity, 
			Print[indent, expr];
			linesAlreadyPrinted++;
			Which[
				ps === Full,
				linesAlreadyPrinted++; 
				Print[indent, "Stack: ", tr],
				ps === Recursive && Count[tr, Print] > 0,
				linesAlreadyPrinted++;
				Print[indent, "Recursive Call: ", tr]
			];
		];
	]

linesAlreadyPrinted = 0;

(* Measures the calculation of an expression and signs it with a time stamp *)	
SetAttributes[QATiming, {HoldAll, SequenceHold}]

QATiming[ex_] := 
	Module[{date, elapsed, res},
		date = DateString[];
		{ elapsed, res } = Timing[ex];
		{ { date, elapsed }, res }
	]

(* Prints out an expression at a certain stage of an evaluation chain and returns the expression unchained *)
QASniff[ex_, descr_String : "Sniffing..."] :=
	Module[{},
		QAPrint[descr, PrtLevel->2];
		QAPrint[ex, PrtLevel->2];
		ex
	]

(*****************************************************************************)

End[]

EndPackage[]
