(* Wolfram Language package *)

(*	*******************************************************************************************************************
	Ideas about new handling of index position in QASimplescript notation
	
	Status:
	- In the moment there is no meaning assigned to upper and lower indices
	- We try to conserve the position as much as possible
	- We try to recognize and preserve the position
	- In the rules we try to enforce an uniform usage
	
	Ideas about newly desired handling:
	- We introduce a setting "IndexCharacter"
	- Three values are possible:
	-- contravariant - express co-variant indices in terms of contravariant ones
	-- covariant - express contra-variant indices in terms of co-variant ones
	-- mixed - preserve the index character
	- For the handling of rules the index character is ignored (as much as possible)

	******************************************************************************************************************* *)


BeginPackage["QuantumAlgebra`RulesSupplement`", 
	{ 
		"QuantumAlgebra`", 
		"QuantumAlgebra`Constants`", 
		"QuantumAlgebra`Analyze`", 
		"QuantumAlgebra`Options`", 
		"QuantumAlgebra`Agreements`" 
	}]


(* ::Section:: *)
(* Exported Symbols *)

IntegrationConstantQ::usage = 
	"IntegrationConstantQ[ex, x, min, max] checks whether an expression ex is an integration constant, i.e. " <>
	"free of an integration or summation variable x scripted by integer indices between min and max. " <>
	"Also the integration/summation variables defined as Explicit Integration Variable play a role here."

VariableQ::usage = 
	"VariableQ[symbol] checks whether an expression is a valid variable."

QAIntegrate::usage = 
	"QAIntegrate[ex, range, dim, var] defines and eventually simplifies an Integration in the context of Quantum Algebra"

QASum::usage = 
	"QASum[ex, var] defines and eventually simplifies an Summation in the context of Quantum Algebra"

QAD::usage = 
	"QAD performs derivation in the context of Quantum Algebra"

DefineCommutatorRule::usage =
	"DefineCommutatorRule[anti, continuous, normal, dim] uses a Commutator / Anti-Commutator to sort operators"
	
DefineNameOrderingRule::usage =
	"DefineNameOrderingRule[anti, dim] uses a Commutator / Anti-Commutator to sort operators according to name order"

QAReplace::usage = 
	"QAReplace[ex, rules] replaces in the expression ex by applying one named set of Replacement Rules. \n" <>
	"QAReplace[ex, { rules }] replaces in the expression ex by applying one or more named sets of Replacement Rules. \n" <>
	"QAReplace[rules][ex] the operator form of the function, can easily be used in evaluation chaines."
	
QAReplaceRepeated::usage = 
	"QAReplaceRepeated[ex, rules] replaces repeatedly in the expression ex by applying one named set of Replacement Rules. \n" <>
	"QAReplaceRepeated[ex, { rules }] replaces repeatedly in an expression by applying one or more named sets of Replacement Rules. \n" <>
	"QAReplaceRepeated[rules][ex] the operator form of the function, can easily be used in evaluation chaines."

QAInitializeReplacementRules::usage = 
	"QAInitializeReplacementRules[] initializes a built-in set of Replacement Rules"

QASimplify::usage = 
	"QASimplify[ex, set, level] performs Simplify using a given named set of Assumptions. \n" <>
	"QASimplify[set][ex] is the operator form of QASimplify."

QAFullSimplify::usage = 
	"QAFullSimplify[ex, set, level] performs FullSimplify using a given named set of Assumptions. \n" <>
	"QAFullSimplify[set][ex] is the operator form of QAFullSimplify."

RowQ::usage =
	"RowQ[matrix] checks whether a Row Matrix is given."

ColumnQ::usage =
	"ColumnQ[matrix] checks whether a Column Matrix is given."

MakeRow::usage =
	"MakeRow[matrix] makes a Row Matrix from a Vector or a Column Matrix."

MakeColumn::usage =
	"MakeColumn[matrix] makes a Column Matrix from a Vector or a Row Matrix."

QASubsuperscript::usage = 
	"QASubsuperscript[symbol, {lower},{upper}] supports enhanced script notation."

QASimplescript::usage = 
	"QASimplescript[symbol, index, pos] supports simple script notation."

QAGetSimplescript::usage = 
	"QAGetSimplescript[symbol, index, pos] supports simple script notation."
	
Factorise::usage =
	"Factorise[ex, pat] performs a hierarchical factorisation in the order of a list of given patterns."	

(* ::Section:: *)
(* Implementation of the Package *)

Begin["`Private`"]


(* ::Subsection:: *)
(* Predicates *)

(* Constant Definition *)
IntegrationConstantQ[ex_, x__, min_Integer, max_Integer] :=
	Module[{symbolsEqualQ, freeExplicitQ},
		symbolsEqualQ[s1_Symbol, s2_Symbol] := Order[SymbolName[s1], SymbolName[s2]] == 0;
		
		freeExplicitQ[ex2_, explicit: Subscript[name_Symbol, idx_Integer], ivar_Symbol] :=
			FreeQ[ex2, explicit] || !symbolsEqualQ[name, ivar];
		freeExplicitQ[ex2_, explicit: Superscript[name_Symbol, idx_Integer], ivar_Symbol] :=
			FreeQ[ex2, explicit] || !symbolsEqualQ[name, ivar];
		
		freeExplicitQ[ex2_, explicit: QASimplescript[name_Symbol, idx_Integer, _], ivar_Symbol] :=
			FreeQ[ex2, explicit] || !symbolsEqualQ[name, ivar];
		
		freeExplicitQ[ex2_, explicit_, ivar_Symbol] :=
			True;
		
		NumericQ[ex] ||
		And@@Flatten[Outer[freeExplicitQ[ex, Subscript[#1, #2], #1]&, { x }, Range[min, max]]] && 
		And@@Flatten[Outer[freeExplicitQ[ex, Superscript[#1, #2], #1]&, { x }, Range[min, max]]] &&
		And@@Flatten[Outer[freeExplicitQ[ex, QASimplescript[#1, #2, _], #1]&, { x }, Range[min, max]]] && 
		And@@Flatten[Outer[freeExplicitQ[ex, #2, #1]&, { x }, QAGetExplicitIntegrationVariables[]]] &&
		!MatchQ[ex, _DiracDelta]
	]


(* Variable Name *)
VariableQ[x_] := MatchQ[x, _Symbol] || MatchQ[x, OverBar[_Symbol]] || MatchQ[x, OverTilde[_Symbol]]

(* ::Subsection:: *)
(* DiscreteDelta / DiracDelta *)

Unprotect[DiscreteDelta]
DiscreteDelta /: DiscreteDelta[args__] ^ 2 := DiscreteDelta[args]
Protect[DiscreteDelta]

Unprotect[DiracDelta]

(*	This rule is mathematically not correct, but yields correct results
	Otherwise we get unresolved DiracDelta sub expressions!
	TODO: find a mathematical reason and algorithm
*)
DiracDelta /: DiracDelta[args__] ^ 2 := DiracDelta[args]

Protect[DiracDelta]

(* ::Subsection:: *)
(* Exponential with I as factor *)

(* Put I outside the brackets as this is required by certain rules *)

Unprotect[Power]

Power /: Power[E, I a_ + I b_] := Power[E, I (a + b)]
Power /: Power[E, I a_ - I b_] := Power[E, I (a - b)]
Power /: Power[E, -I a_ - I b_] := Power[E, -I (a + b)]

Protect[Power]


(* ::Subsection:: *)
(* Integration Rules *)

(* Integration about Zero yields Zero *)
QAIntegrate[0, v: (_Symbol | Infinity), _Integer, __Symbol] := 0


QAIntegrate[0., v: (_Symbol | Infinity), _Integer, __Symbol] := 0


(* Integration about Constant is Constant times integration about one *)
QAIntegrate[c_, v: (_Symbol | Infinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] && ! c === 1 :=
	c QAIntegrate[1, v, n, x]
	
	 
(* Ideas:
   - Length of integration variables' sequence irrelevant
   - v is common area
   - v as Symbol cannot be Infinity
   
   Integration about one is Integration Area 
*)
QAIntegrate[1, v_Symbol, n_Integer, x__Symbol] /;
	! v === Infinity := 
	v


(* Integration about Constant times rest integrand *)
QAIntegrate[c_ p_, v: (_Symbol | Infinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	c QAIntegrate[p, v, n, x]
	
	
(* Ideas:
   - for the multiplication of integrals we need a cross product of the ranges
   - in case of Infinity we can symbolically replace Infinity x Infinity by Infinity
   
   Integration times Integration with infinite integration area 
*)
QAIntegrate /: QAIntegrate[p_, Infinity, n_Integer, x__Symbol] QAIntegrate[q_, Infinity, n_Integer, y__Symbol] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p q, Infinity, n, Sequence[x, y]]


(* Integration times Integration with finite integration areas *)
QAIntegrate /: QAIntegrate[p_, u: _Symbol, n_Integer, x__Symbol] QAIntegrate[q_, v: _Symbol, n_Integer, y__Symbol] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p q, AngleBracket[u, v], n, Sequence[x, y]]


(* New Algorithm uses Solve and avoids tricky sign determination
   2019-04-29
*)

(* Integration about 2 integration variables where DiracDelta is in integrand containing both of them *)
QAIntegrate[exp: Times[facs__, dd: DiracDelta[fa_, la___]], v: Infinity, n_Integer, Sequence[a_Symbol, b_Symbol]] /;
	Module[{},
		ruleMatchesQ[{fa, la}, a, b] && FreeQ[{ facs }, _DiracDelta]
	] :=
	Module[{facsrepl},
		facsrepl = Sequence@@({facs} /. getReplacement[{fa, la}, a, b]);
		QAIntegrate[Times[facsrepl], v, n, a]
	]


(* Integration about 1 integration variable where DiracDelta is used
   REMARK: uses arbitrary index positions
   2019-04-23 
   
   Integration about 1 integration variable with DiracDelta in integrand 
*)
QAIntegrate[exp: Times[facs__, dd: DiracDelta[fa_, la___]], v: Infinity, n_Integer, a_Symbol] /;
	Module[{},
		ruleMatchesQ[{fa, la}, a]
	] :=
	Module[{facsrepl},
		facsrepl = Sequence@@({facs} /. getReplacement[{fa, la}, a]);
		Times[facsrepl]
	]


(* Integrate about an Exponential resolved to DiracDelta / DiscreteDelta
   New simplified rule, substituting the older ones
   2019-04-28
   
   critic 2020-03-30:
   - does not work for n = 1 without time item
   - cannot detect error
   - we try to generalize with 2 additional factors 2I, -2I -> what about normalization?
   
   tests after changes:
   - Klein Gordon working
   - Dirac working
   - Maxwell like before : DiracDelta[0, 0, 0] generated
*)
QAIntegrate[Power[E, (fac: (I | -I | 2 I | -2 I)) * exp_], v: (_Symbol | Infinity), n_Integer, x_Symbol] /;
	Module[{lst, xvec},
		lst = List@@(exp // FullSimplify);
		xvec = Cases[lst, Times[QASimplescript[x, i_, pos_], __] -> QASimplescript[x, i, pos]];
		Piecewise[
			{
				{ MatchQ[xvec, Table[QASimplescript[x, i, pos_], {i, 0, n}]], Length[xvec] == n + 1 },
				{ MatchQ[xvec, Table[QASimplescript[x, i, pos_], {i, 1, n}]], Length[xvec] == n }
			},
			False
		]
	] :=
	Module[{lst, sel, drop, delta, norm, time},
		drop[ls_List] := If[FreeQ[ls, QASimplescript[x, 0, _]], 0, 1];
		delta[] := If[! v === Infinity, DiscreteDelta, DiracDelta];
		norm[] := If[! v === Infinity, v, (2 * Pi) ^ n];
		time[ls_List] := If[FreeQ[ls, QASimplescript[x, 0, _]], 1, Power[E, fac * ls[[1]]]]; 
		
		lst = List@@(exp // FullSimplify);
		sel = Cases[lst, Times[QASimplescript[x, i_, _], k_, l___] -> Times @@ {k, l}];
		delta[] @@ Drop[sel, drop[lst]] * time[lst] * norm[]
	]


(* 2019-05-23: Maybe we had an equivalent rule, but it was lost *)
(* Special case of the orthogonality of Fourier Eigen solutions -> generalize? 

   critic:
   - cannot recognize background
   - is this really correct and complete?
   - see rule above

QAIntegrate[Power[E, (fac: (2 I | -2 I)) * exp_], v_Symbol, n_Integer, x_Symbol] /; 
	Module[{lst, xvec},
		lst = List@@(exp // FullSimplify);
		xvec = Cases[lst, Times[QASimplescript[x, i_, pos_], __] -> QASimplescript[x, i, pos]];
		Piecewise[
			{
				{ MatchQ[xvec, Table[QASimplescript[x, i, pos_], {i, 0, n}]], Length[xvec] == n + 1 },
				{ MatchQ[xvec, Table[QASimplescript[x, i, pos_], {i, 1, n}]], Length[xvec] == n }
			},
			False
		]
	] :=
	0	
*)


(* Simplification for skew-symmetric integrand expression *)
(* TODO: Integrand can be anything, complicated expressions as well -> perform rule at desired times

*)
QAIntegrate[ex_, Infinity, dim_Integer, k_Symbol] /;
	FreeQ[ex, Operator] &&
	(skewSymmetricQ[ex, k, dim, True] || skewSymmetricQ[ex, k, dim, False]) :=
	0
	
	
(* ::Subsection:: *)
(* Summation Rules *)

(* Summation about Zero yields Zero *)
QASum[0, __Symbol] := 0

QASum[0., __Symbol] := 0


(* Summation about a Constant *)
QASum[c_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] && ! c === 1 :=
	c QASum[1, x] 


(* Summation about Constant times rest expression *)
QASum[c_ p_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	c QASum[p, x]


(* Product of summations *)
QASum /: QASum[p_, x__Symbol] QASum[q_, y__Symbol] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QASum[p q, Sequence[x, y]]


(* Summation about 2 summation variables where DiscreteDelta is used containing both of them *) 
QASum[exp: Times[facs__, dd: DiscreteDelta[fa_, la___]], Sequence[a_Symbol, b_Symbol]] /;
	Module[{},
		ruleMatchesQ[{fa, la}, a, b] 
	] :=
	Module[{facsrepl},
		facsrepl = Sequence@@({facs} /. getReplacement[{fa, la}, a, b]);
		QASum[Times[facsrepl], a]
	]


(* Simplification for anti-symmetric summation expression, dimension hard coded
   See QAIntegrate above. 

*)
QASum[ex_, k_Symbol] /;
	FreeQ[ex, Operator] &&
	(skewSymmetricQ[ex, k, 3, True] || skewSymmetricQ[ex, k, 3, False]) :=
	0


(* Rule resolves a Piecewise construct *)
Unprotect[Piecewise]

Piecewise::fmt = "Illegal format of Piecewise expression"

Piecewise /: (ex1: Piecewise[ex: _List, 0]) /;
	! FreeQ[ex, _QASum] :=
	Module[{ex2, cond, svars, vars, rules, up, res},
		{ ex2, cond } = First[ex];
		svars = Cases[ex2, QASum[_, a_Symbol, b_Symbol] -> {b, a}, Infinity, 1];
		If[vars == {}, 
			Message[Piecewise::fmt]
		];																			(* Error Message *)
		svars = svars[[1]]; 														(* yields list of summation variables { b, a } *)
		vars = Cases[cond, QASimplescript[svars[[1]], _, _], 5]; 					(* yields list of indexed variables to be replaced *)
		up = Cases[cond, QASimplescript[svars[[1]], _, up2_] -> up2, 5, 1][[1]]; 	(* yields index position *)
		rules = Solve[cond, vars];
		If[rules == {}, 
			Message[Piecewise::fmt]
		];																			(* Error Message *)
		rules = Append[First[rules], Apply[QASimplescript[#1, 0, up] -> QASimplescript[#2, 0, up]&, svars]];
		res = ex2 /. rules;
		res = res /. { Apply[QASum[ex3_, #2, #1] -> QASum[ex3, #2]&, svars] }
	]

Protect[Piecewise]


(* ::Subsection:: *)
(* Exchange order of Integration / Summation *)
QAIntegrate[QASum[ex_, svar__Symbol], v: (_Symbol | Infinity), n_Integer, ivar__Symbol] := QASum[QAIntegrate[ex, v, n, ivar], svar]


(* ::Subsection:: *)
(* Integration Rules with Operators *)

(* Integration about operator product where one is a Constant *)
QAIntegrate[c_ ** p_, v: (_Symbol | Infinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	c ** QAIntegrate[p, v, n, x]


(* Integration about operator product where one is a Constant *)
QAIntegrate[p_ ** c_, v: (_Symbol | Infinity), n_Integer, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, n] :=
	QAIntegrate[p, v, n, x] ** c


(* Hermitian of an Integral *)
Hermitian[QAIntegrate[ex_, v: (_Symbol | Infinity), n_Integer, x__Symbol]] := QAIntegrate[Hermitian[ex], v, n, x]


(* Product of Integrals both containing operators with finite integration areas *)
QAIntegrate /: QAIntegrate[p_, u: _Symbol, n_Integer, x__?VariableQ] ** QAIntegrate[q_, v: _Symbol, n_Integer, y__?VariableQ] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p ** q, AngleBracket[u, v], n, Sequence@@Sort[List[x, y]]]


(* Product of Integrals both containing operators with infinite integration areas *)
QAIntegrate /: QAIntegrate[p_, v: Infinity, n_Integer, x__?VariableQ] ** QAIntegrate[q_, v: Infinity, n_Integer, y__?VariableQ] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QAIntegrate[p ** q, v, n, Sequence@@Sort[List[x, y]]]


(*
	We want to establish a simplification rule for Integration / Summation (QAIntegrate / QASum) about the whole
	integration / summation range
	
	The rule is as follows:
	QAIntegrate[f[k], Infinity, dim, k] := 0 if f[-k] == -f[k], e.g. f is an antisymmetric function related to its
	(indexed) parameter k.
	
	Problem:
	To check, we must replace k -> -k. But this changes the order in an operator product. Maybe an automatic rule
	is applied which changes the sign back !!
	
	Solution:
	Serve only a certain pattern with an operator product

 	Integration about operator product and Skew Symmetric rest integrand 
*)
QAIntegrate[Times[facs1__, Operator[a_Symbol, seq___, lst1: {fa_, la___}, r: {___}] ** Operator[a_Symbol, seq___, lst2: {__}, s: {___}]], Infinity, dim_Integer, k_Symbol] /;
	Module[{i},
		MatchQ[lst1, Table[QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		MatchQ[lst2, Table[-QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		r == s &&
		({ seq } == {} || seq == "H") &&
		skewSymmetricQ[Times[facs1], k, Length[lst1], upperQ[fa]]
	] :=
	0


(* Eliminate double-negative indices in operator product *)
(* 2019-06-05 *)
QAIntegrate[int: Times[facs1__, Operator[a_Symbol, seq1___, lst1: {-fa_, ___}, r: {___}] ** Operator[b_Symbol, seq2___, lst2: {__}, s: {___}]], Infinity, dim_Integer, k_Symbol, l_Symbol] /;
	Module[{ },
		((indicesNegative[lst1, k, upperQ[fa], dim] && indicesNegative[lst2, l, upperQ[fa], dim]) ||
		(indicesNegative[lst2, k, upperQ[fa], dim] && indicesNegative[lst1, l, upperQ[fa], dim])) &&
		adjunctValidQ[seq1] &&
		adjunctValidQ[seq2]
	] :=
	Module[{newint},
		newint = int;
		newint = newint /. replaceNegative[k, upperQ[fa], dim];
		newint = newint /. replaceNegative[l, upperQ[fa], dim];
		QAIntegrate[newint, Infinity, dim, k, l]
	]


(* Eliminate double-negative indices in operator product *)
QAIntegrate[int: Times[facs1__, Operator[a_Symbol, seq1___, lst1: {-fa_, ___}, r: {___}] ** Operator[b_Symbol, seq2___, lst2: {__}, s: {___}]], Infinity, dim_Integer, k_Symbol] /;
	Module[{ },
		indicesNegative[lst1, k, upperQ[fa], dim] && indicesNegative[lst2, k, upperQ[fa], dim] &&
		adjunctValidQ[seq1] &&
		adjunctValidQ[seq2]
	] :=
	Module[{newint},
		newint = int;
		newint = newint /. replaceNegative[k, upperQ[fa], dim];
		QAIntegrate[newint, Infinity, dim, k]
	]


(* Sort negative indices in operator product
   Desired order : negative indices on right-hand side 
*)
QAIntegrate[Times[facs1__, Operator[a_Symbol, seq___, lst1: {__}, r: {___}] ** Operator[b_Symbol, seq___, lst2: {fa_, la___}, s: {___}]], Infinity, dim_Integer, k_Symbol] /;
	Module[{i},
		MatchQ[lst1, Table[-QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		MatchQ[lst2, Table[QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		({ seq } == {} || seq == "H")
	] :=
	Module[{},
		QAIntegrate[Times[
			Sequence@@({ facs1 } /. QASimplescript[k, i_, upperQ[fa]] /; 1 <= i <= Length[lst1] -> -QASimplescript[k, i, upperQ[fa]]),
			Operator[a, seq, lst2, r] ** Operator[b, seq, lst1, s]],
			Infinity, dim, k]
	]


(* ::Subsection:: *)
(* Summation Rules with Operators *)

(* Summation about operator product where one operator is Constant *)
QASum[c_ ** p_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	c ** QASum[p, x]


(* Summation about operator product where one operator is Constant *)
QASum[p_ ** c_, x__Symbol] /;
	IntegrationConstantQ[c, x, 1, 3] :=
	QASum[p, x] ** c


(* Hermitian of Sum *)
Hermitian[QASum[ex_, x__Symbol]] := QASum[Hermitian[ex], x]


(* Non-commutative product of Sums *)
QASum /: QASum[p_, x__?VariableQ] ** QASum[q_, y__?VariableQ] /;
	Intersection[Map[SymbolName, {x}], Map[SymbolName, {y}]] == {} :=
	QASum[p ** q, Sequence@@Sort[List[x, y]]]


(* Simplification for anti-symmetric summation expression *)
QASum[Times[facs1__, Operator[a_Symbol, seq___, lst1: {fa_, la___}, r: {___}] ** Operator[a_Symbol, seq___, lst2: {__}, s: {___}]], k_Symbol] /;
	Module[{i},
		MatchQ[lst1, Table[QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		MatchQ[lst2, Table[-QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		r == s &&
		({ seq } == {} || seq == "H") &&
		skewSymmetricQ[Times[facs1], k, Length[lst1], upperQ[fa]]
	] :=
	0


(* Eliminate double-negative indices in operator product *)
QASum[int: Times[facs1__, Operator[a_Symbol, seq1___, lst1: {-fa_, ___}, r: {___}] ** Operator[b_Symbol, seq2___, lst2: {__}, s: {___}]], k_Symbol, l_Symbol] /;
	Module[{ },
		((indicesNegative[lst1, k, upperQ[fa], Length[lst1]] && indicesNegative[lst2, l, upperQ[fa], Length[lst1]]) ||
		(indicesNegative[lst2, k, upperQ[fa], Length[lst1]] && indicesNegative[lst1, l, upperQ[fa], Length[lst1]])) &&
		adjunctValidQ[seq1] &&
		adjunctValidQ[seq2]
	] :=
	Module[{newint},
		newint = int;
		newint = newint /. replaceNegative[k, upperQ[fa], Length[lst1]];
		newint = newint /. replaceNegative[l, upperQ[fa], Length[lst1]];
		QASum[newint, k, l]
	]


(* Eliminate double-negative indices in operator product *)
QASum[int: Times[facs1__, Operator[a_Symbol, seq1___, lst1: {-fa_, ___}, r: {___}] ** Operator[b_Symbol, seq2___, lst2: {__}, s: {___}]], k_Symbol] /;
	Module[{ },
		indicesNegative[lst1, k, upperQ[fa], Length[lst1]] && 
		indicesNegative[lst2, k, upperQ[fa], Length[lst1]] &&
		adjunctValidQ[seq1] &&
		adjunctValidQ[seq2]
	] :=
	Module[{newint},
		newint = int;
		newint = newint /. replaceNegative[k, upperQ[fa], Length[lst1]];
		QASum[newint, k]
	]
	

(* Sort negative indices in operator product
   Desired order : negative indices on right-hand side 
*)
QASum[Times[facs1__, Operator[a_Symbol, seq___, lst1: {__}, r: {___}] ** Operator[b_Symbol, seq___, lst2: {fa_, la___}, s: {___}]], k_Symbol] /;
	Module[{i},
		MatchQ[lst1, Table[-QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		MatchQ[lst2, Table[QASimplescript[k, i, upperQ[fa]], {i, 1, Length[lst1]}]] &&
		({ seq } == {} || seq == "H")
	]:=
	Module[{},
		QASum[Times[
			Sequence@@({ facs1 } /. QASimplescript[k, i_, upperQ[fa]] /; 1 <= i <= Length[lst1] -> -QASimplescript[k, i, upperQ[fa]]),
			Operator[a, seq, lst2, r] ** Operator[b, seq, lst1, s]],
			k]
	]


(* ::Subsection:: *)
(* Derivation Rules *)

SetAttributes[QAD, { Listable}]

(* Derivation variable with factor -1 *)
QAD[ex_, -1 var_] := -QAD[ex, var]


(* Derive Integral *)
QAD[QAIntegrate[ex_, v: (_Symbol | Infinity), n_Integer, x__Symbol], seq__] := QAIntegrate[QAD[ex, seq], v, n, x]


(* Derive Sum *)
QAD[QASum[ex_, x__Symbol], seq__] := QASum[QAD[ex, seq], x]


(* Derive Sum *)
QAD[fac1_ + fac2_, seq__] := QAD[fac1, seq] + QAD[fac2, seq]


(* Derive commutative product *)
QAD[fac1_ * fac2_, var_] := QAD[fac1, var] fac2 + fac1 QAD[fac2, var]


(* Derive non-commutative product *)
QAD[fac1_ ** fac2_, var_] := QAD[fac1, var] ** fac2 + fac1 ** QAD[fac2, var]


(* Derive constant *)
QAD[c_, var_] /; FreeQ[c, var] := 0


(* Delegate to built-in Derivative *)
QAD[ex_, seq__] /; 
	! Head[ex] === QAIntegrate && 
	! Head[ex] === QASum && 
	! Head[ex] === Plus && 
	! Head[ex] === Times && 
	! Head[ex] === NonCommutativeMultiply := 
	D[ex, seq]


(* ::Subsection:: *)
(* Matrix Operations *)

(* Matrix operations as used in Spinor Theory *)

(* Checks for a Row Matrix *)
RowQ[m: {{__}} ? MatrixQ] := True

RowQ[m_List ? MatrixQ] := False


(* Checks for a Column Matrix *)
ColumnQ[m_List ? MatrixQ] := RowQ[Transpose[m]]


(* Makes a Row Matrix for either a Vector or a Column Matrix *)
MakeRow[m_List] /; VectorQ[m] := { m }

MakeRow[m_List] /; ColumnQ[m] := Transpose[m]

MakeRow[m_List] := m


(* Makes a Column Matrix from either a Vector or a Row Matrix *)
MakeColumn[m_List] /; VectorQ[m] := Transpose[MakeRow[m]]

MakeColumn[m_List] /; RowQ[m] := Transpose[m]

MakeColumn[m_List] := m


(* The Hermitian operation on a Matrix or a Vector
   - Matrix: apply Hermitian on each element and transpose
   - Vector: apply Hermitian on each element
*)
Hermitian[m_List] /; MatrixQ[m] := Map[Hermitian, m // Transpose, {2}]

Hermitian[m_List] /; VectorQ[m] := Map[Hermitian, m]


(* A scalar list Times an operator *)
Operator /: c_List op_Operator := Map[# * op&, c]


Unprotect[NonCommutativeMultiply];

(* The Distribution Law for NonCommutativeMultiply,
   provided summands contain QASum expressions 
*)
(smd1_ + smd2:__) ** sum_Plus /; !FreeQ[smd1, _QASum] := smd1 ** sum + (Plus@@{ smd2 }) ** sum

sum_Plus ** (smd1_ + smd2:__) /; !FreeQ[smd1, _QASum] := sum ** smd1 + sum ** (Plus@@{ smd2 })


(* The Distribution Law for NonCommutativeMultiply with List,
   provided summands contain QASum expressions 
*)
(smd1_ + smd2:__) ** mat_List /; !FreeQ[smd1, _QASum] := smd1 ** mat + (Plus@@{ smd2 }) ** mat

mat_List ** (smd1_ + smd2:__) /; !FreeQ[smd1, _QASum] := mat ** smd1 + mat ** (Plus@@{ smd2 })


(* Another form of Distribution Law for NonCommutativeMultiply between List and QASum of Lists *)
(* 2019-06-11 TODO: corresponding rule with Integrate *)
QASum[mat1_List, seq__] ** mat2_List := QASum[mat1 ** mat2, seq]

mat2_List ** QASum[mat1_List, seq__] := QASum[mat2 ** mat1, seq]


(* Matrix multiplication with matrices containing Operator(s)
   TODO: For these rules to become active, Operators are required -> generalize? 
*)
mat1_List ? RowQ ** mat2_List ? VectorQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	mat1 ** MakeColumn[mat2]

mat1_List ? VectorQ ** mat2_List ? ColumnQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	MakeRow[mat1] ** mat2
	
mat1_List ? RowQ ** mat2_List ? ColumnQ /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	Inner[NonCommutativeMultiply, mat1, mat2, Plus][[1, 1]]
	
mat1_List ** mat2_List /; 
	!FreeQ[mat1, _Operator] || !FreeQ[mat2, _Operator] := 
	Inner[NonCommutativeMultiply, mat1, mat2, Plus]
	

mat1_List ? RowQ ** mat2_List ? ColumnQ /; 
	FreeQ[mat1, _Operator] && FreeQ[mat2, _Operator] := 
	Inner[Times, mat1, mat2, Plus][[1, 1]]
	
mat1_List ** mat2_List /; 
	FreeQ[mat1, _Operator] && FreeQ[mat2, _Operator] := 
	Inner[Times, mat1, mat2, Plus]


(* Outer Product of Vectors *)
vec1_List ? VectorQ \[SmallCircle] vec2_List ? VectorQ :=
	Outer[NonCommutativeMultiply, vec1, vec2]
	 

(* NonCommutativeMultiply reduced to Times *)
fac1_ ** fac2_ /; QuantumAlgebra`Private`SimpleConstantQ[fac1] && QuantumAlgebra`Private`SimpleConstantQ[fac2] := fac1 * fac2 * Operator[1]

fac1_ ** fac2_ /; QuantumAlgebra`Private`SimpleConstantQ[fac1] || QuantumAlgebra`Private`SimpleConstantQ[fac2] := fac1 * fac2

Protect[NonCommutativeMultiply];


(* ::Subsection:: *)
(* Functions Generate Replacement Rules for (Anti-) Commutators *)

(*	New Design Of Commutator Rule
	1. Upper parameter removed, obsolete
	2. Originate the rule from Commutation Relation, solve equation
	3. Can we save the "reverse ordering" case?
*)
DefineCommutatorRule[anti_, continuous_, normal_, dim_Integer : 3] :=
	Module[{zeta, antisign, head, list1, list2, op1, op2, part1, part2, comm, commrel, rule1, rule2, a, b, h, k, r, s, upper, kdph},
		zeta[r1_Integer : 1] := If[r1 == 0, -1, +1];
		antisign = If[anti, -1, +1];
		head = If[continuous, DiracDelta, DiscreteDelta];
		list1 = Table[QASimplescript[h, i, upper], {i, 1, dim}];
		list2 = Table[QASimplescript[k, i, upper], {i, 1, dim}];
		op1 = Operator[a, list1, { r }];
		op2 = Operator[b, "H", list2, { s }];
		comm = head@@Inner[#1 - #2&, list2, list1, List] KroneckerDelta[r, s] zeta[r] Operator[1] kdph;
		commrel = ((op1 ** op2 - antisign op2 ** op1) == comm);
		If[normal,
			part1 = (op1 ** op2),
			part1 = (op2 ** op1)
		];
		part2 = Solve[commrel, part1][[1, 1, 2]];
		part1 = part1 /. Map[# -> Pattern[#, Blank[]]&, {a, b, h, k, upper}];
		part1 = part1 /. Map[# -> Pattern[#, BlankNullSequence[]]&, {r, s}];
		
		rule1 = Apply[ HoldPattern[#1] :> #2 /; Order[#3, #4] == 0&, { part1, part2 /. kdph -> 1, a, b }];
		rule2 = Apply[ HoldPattern[#1] :> #2 /; Order[#3, #4] != 0&, { part1, part2 /. kdph -> 0, a, b }];
		{ rule1, rule2 }
	]


(* 	Used to bring Operators in unique (canonical) order
	In the case of (Anti-) Commutating operators
	Order test improved: use the built-in function for canonical order
*)
DefineNameOrderingRule[anti_, dim_Integer : 3] :=
	Module[{sign, list1, list2, op1, op2, op2repl, part1, a, b, h, k, r, s, n, seq, sn, up1, up2, res, indexRule, ruleParams},
		sn[sy_Symbol, i2_Integer] := Symbol[SymbolName[sy] <> ToString[i2]];
		sign = If[anti, -1, +1];
		list1 = Table[QASimplescript[k, i, up1], {i, 1, dim}];
		list2 = Table[QASimplescript[h, i, up2], {i, 1, dim}];
		op1 = Operator[a, seq, list1, r];
		op2 = Operator[b, seq, list2, s];
		op2repl = Operator[b, seq, Table[sn[n, i], {i, 1, dim}], s];
		part1 = op1 ** op2;
		part1 = part1 /. Map[# -> Pattern[#, Blank[]]&, {a, b, k, up1}];
		part1 = part1 /. Map[QASimplescript[h, #, up2] -> Pattern[sn[n, #] // Evaluate, QASimplescript[h_, #, up2_] | -QASimplescript[h_, #, up2_]]&, Range[dim]];
		part1 = part1 /. seq -> seq___;
		part1 = part1 /. Map[# -> Pattern[#, List[BlankNullSequence[]]]&, {r, s}];

		res = sign op2repl ** op1;
		indexRule = -QASimplescript[a1_, b1_, c1_] :> QASimplescript[a1, b1, c1];
		ruleParams = { part1, res, op1, op2repl, indexRule };
		Apply[HoldPattern[#1] :> #2 /; Order[#3, #4 /. #5] < 0&, ruleParams]
	]


(*	Factorisation
	performs a hierarchical factorisation in the order of a list of given patterns
	one pattern must not contain more than one factor
	ex 	- expression to be factorised
	pat - List of patterns
*)
Factorise[ex_, pat_List] :=
	Module[{},
		If[pat == {},
			ex,
			Factorise[Factorise[ex, First[pat]], Rest[pat]] 
		]
	]

Factorise[ex_, pat_] :=
	ex /. { sum_Plus /;
		Module[{lst, res},
			lst = List@@sum;
			res = And@@Map[MatchQ[#, Times[pat, ___]]&, lst];
			(* If[res, QAPrint["Factorise Check: ", ex, " : ", pat, " : ", sum], False]; *)
			res
		] :>
		Module[{res, dummy},
			QAPrint["Factorise: ", pat];
			res = Cases[sum, Times[pa: pat, x__] -> {pa, dummy[x]}];
			res = Gather[res, Order[#1[[1]], #2[[1]]] == 0&];
			res = res /. dummy -> Times;
			Plus@@Map[Times[#[[1, 1]], Plus@@#[[1 ;; , 2]]]&, res]
		]}
	
	
(* ::Subsection:: *)
(* Reverse Application of the Distribution Law for Operators or Operator Products with scalar factors *)

Unprotect[Plus]

(*	TODO: ATTENTION! This works against FullDistributionLaw -> Better use DistributionLaw instead.
*)	

Plus /:
	c_ a_Operator ** b_Operator + d_ a_Operator ** b_Operator :=
	(c + d) a ** b 

Plus /:
	c_ a_Operator + d_ a_Operator :=
	(c + d) a
	
Plus /:
	c_ x_DiracDelta ^ 2 + d_ x_DiracDelta ^ 2 :=
	(c + d) x ^ 2 

Protect[Plus]

Unprotect[Times]

Times /:
	Operator[1] facs__ := Times[facs] /; !FreeQ[{ facs }, _Operator]

Protect[Times]


(* ::Subsection:: *)
(* Replacement Rules *)

(* A conditional replacement using rule
   Only if the matched pattern is not contained in the Exponent of the E - function, a transformation
   is performed. 
*)
QAReplace[ex_, rule_RuleDelayed] :=
	Module[{exponents, powerCaseQ, cases, critical, transformed},
		exponents[] := Cases[ex, Power[E, expo_] -> expo, Infinity]; 
		powerCaseQ[case_] := !FreeQ[exponents[], case];

		cases = Cases[{ ex }, rule[[1, 1]], Infinity];
		critical = Select[cases, !powerCaseQ[#]&];  (* FreeQ[ex, Power[E, #]]& *)
		transformed = critical /. rule;
		ex /. Inner[Rule, critical, transformed, List]
	]


(* Executes Replacement Rule(s) given as named set *)
QAReplace[ex_, rules_String] :=
	Module[{res, time},
		QAPrint["Replace of rule(s) ", rules];
		{ time, res } = 
			If[rules == "DistributionLaw" || rules == "FullDistributionLaw",
				QAReplace[ex, QAGetReplacementRules[rules][[1]]], 
				ex /. QAGetReplacementRules[rules]
			] // Timing;
		QAPrint[StringTemplate["Took `1`s."][time]];
		res
	]

(* Executes Replacement Rule(s) given as list of named set(s) *)
QAReplace[ex_, rules: { __String }] := 
	Module[{},
		QAPrint["Replace of set of rule(s) ", rules];
		ex /. Dispatch[Flatten[Map[QAGetReplacementRules[#]&, rules]]]
	]
	
(* Executes Replacement Rule(s), operator form *)
QAReplace[rules: _String | { __String }][ex_] := QAReplace[ex, rules] 


(* Conditional Replace Repeated *)
QAReplaceRepeated[ex_, rule_RuleDelayed] :=
	Module[{oldex, newex},
		oldex = ex;
		While[True,
			QAPrint["In QAReplaceRepeated for FullDistributionLaw", PrtLevel->3];
			newex = QAReplace[oldex, rule];
			If[Order[newex, oldex] == 0, Break[]];			
			oldex = newex;
		];
		newex
	]

(* Executes Replacement Rule(s), given as named set, repeatedly *)
QAReplaceRepeated[ex_, rules_String] := 
	Module[{res, time},
		QAPrint["ReplaceRepeated of rule(s) ", rules];		
		{ time, res } = 
			If[rules == "DistributionLaw" || rules == "FullDistributionLaw",
				QAReplaceRepeated[ex, QAGetReplacementRules[rules][[1]]], 
				ex //. QAGetReplacementRules[rules]
			] // Timing;
		QAPrint[StringTemplate["Took `1`s."][time]];
		res
	]

(* Executes Replacement Rule(s), given as list of named set(s), repeatedly *)
QAReplaceRepeated[ex_, rules: { __String }] :=
	Module[{},
		QAPrint["ReplaceRepeated of set of rule(s) ", rules];
 		ex //. Dispatch[Flatten[Map[QAGetReplacementRules[#]&, rules]]]
	]
	
(* Executes Replacement Rule(s), given as list of named set(s), repeatedly. Operator form. *)
QAReplaceRepeated[rules: _String | { __String }][ex_] := QAReplaceRepeated[ex, rules] 


(* Fixed set of replacement rules. This set will be activated together with Quantum Algebra. *)
QAInitializeReplacementRules[] :=
	Module[{},
		(* Integrate about Sums (Plus operation) is splitted *)
		QAAddReplacementRules[
			{
				HoldPattern[QAIntegrate[Times[f___ : 1, Plus[p__]], v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]] :> 
					Plus @@ Map[QAIntegrate[Times[f, #], v, n, x]&, { p }]
			},
			"IntegrateSplit"];

		(* Sum of Integrates is joined *)
		QAAddReplacementRules[
			{
				HoldPattern[ex_] :> join[ex]
			},
			"IntegrateJoin"];
		
		(* Sum about Sums (Plus operation) is splitted *)
		QAAddReplacementRules[
			{
				(*	TEST NEW VERSION
				HoldPattern[QASum[Times[f___ : 1, p_ + q_], x__Symbol]] :> (QASum[Times[f, p], x] + QASum[Times[f, q], x])
				*)			
				HoldPattern[QASum[Times[f___ : 1, Plus[p__]], x__Symbol]] :> 
					Plus @@ Map[QASum[Times[f, #], x]&, { p }]
			},
			"SumSplit"];
		
		(* Sum of Sums is joined *)
		QAAddReplacementRules[
			{
				(*	TEST NEW VERSION
				HoldPattern[Times[c___ : 1, QASum[p_, x__Symbol]] + Times[d___ : 1, QASum[q_, x__Symbol]]] :> 
					QASum[Times[c, p] + Times[d, q], x]
				*)
				HoldPattern[ex_] :> join[ex, False]
			},
			"SumJoin"];
		
		(* The normal distribution law *)
		QAAddReplacementRules[
			{
				HoldPattern[Times[fac__, sum_Plus]] :> 
					Plus@@Map[Times[fac, #]&, List@@sum] /; 
					FreeQ[{ fac }, _Operator ** _Operator | _Operator]
			},
			"DistributionLaw"];
		
		(* The normal distribution law *)
		QAAddReplacementRules[
			{
				HoldPattern[Times[fac__, sum_Plus]] :> 
					Module[{res},
						res = Plus@@Map[Times[fac, #]&, List@@sum];
						res
					] /;
					FreeQ[{ fac }, Operator[1]]
			},
			"FullDistributionLaw"];
			
		(* The reversed distribution law *)
		QAAddReplacementRules[
			{
				HoldPattern[ex: _Plus] :>
					Module[{ lst, cmn, diverse },
						lst = Apply[List, ex, {0, 1}];
						cmn = Intersection[Sequence@@lst];
						diverse = Map[Complement[#, cmn]&, lst];
						Times[Sequence @@ cmn, Plus @@ (Apply[Times, diverse, 1])]
					] /;
					And@@Map[MatchQ[#, Times[a__, _]]&, List@@ex]
			},
			"ReverseDistributionLaw"];
		
		(* Order of variables in certain constructs like Integrate, Sum, DiracDelta, DiscreteDelta is sorted by Variable Name *)
		QAAddReplacementRules[
			{
				HoldPattern[QAIntegrate[ex_, v: (_Symbol | _DirectedInfinity), n_Integer, seq__?VariableQ]] :> 
					QAIntegrate[ex, v, n, Sequence@@Map[Symbol, Sort[Map[SymbolName, {seq}]]]],
				
				HoldPattern[QASum[ex_, seq__?VariableQ]] :> 
					QASum[ex, Sequence@@Map[Symbol, Sort[Map[SymbolName, {seq}]]]],
				
				defineSortVariablesRule[True],
				
				defineSortVariablesRule[False]
			},
			"SortVariables"];
	True]


(* ::Subsection:: *)
(* Assumption handling and simplification *)

(* Simplification using a given set of assumptions *)
QASimplify[ex_, set_, level_Integer : 1] := 
	Module[{},
		QAPrint[StringTemplate["Simplify with assumptions from `1`."][If[MatchQ[set, _String], set, "default"]], PrtLevel->level];
		Simplify[ex, 
			Assumptions -> Map[And, QAGetAssumptions[set]], 
			TimeConstraint -> QAGetTimeConstraint[set]
		]
	]

(* Simplification using a given set of assumptions, usable as operator *)
QASimplify[set_String : Null][ex_] := QASimplify[ex, set]

(* Full simplification using a given set of assumptions *)
QAFullSimplify[ex_, set_, level_Integer : 1] := 
	Module[{},
		QAPrint[StringTemplate["FullSimplify with assumptions from `1`."][If[MatchQ[set, _String], set, "default"]], PrtLevel->level];
		FullSimplify[ex, 
			Assumptions -> Map[And, QAGetAssumptions[set]], 
			TimeConstraint -> QAGetTimeConstraint[set]
		]
	]

(* Full simplification using a given set of assumptions, usable as operator *)
QAFullSimplify[set_String : Null][ex_] := QAFullSimplify[ex, set]


(* ::Subsection:: *)
(* Default set and Index Character handling *)

(* This rule implements the Covariant / Contravariant / Mixed behavior *)
QAGetSimplescript[symbol_Symbol, idx_Integer, up: (True | False), form_ : StandardForm] :=
	Module[{character, vars},
		{ character, vars } = QAGetIndexCharacter[];
		If[form === InputForm || 
			character == "Mixed" || 
			up == (character == "Contravariant") ||
			! MemberQ[vars, symbol],
			
			(QAPrint[StringTemplate["Leaving Idx `1` for `2`."][idx, symbol], PrtLevel->2];
			QASimplescript[symbol, idx, up]),
			
			(QAPrint[StringTemplate["Changing Idx `1` for `2` to Contravariant."][idx, symbol], PrtLevel->2];
			If[idx == 0, +1, -1] QASimplescript[symbol, idx, !up])
		]
	]

(* Fallback rule *)	
QAGetSimplescript[symbol_, idx_, up_] := QASimplescript[symbol, idx, up]


(* ::Subsection:: *)
(* Internal Rules and Data *)

(* Determines if first argument of a DiracDelta/DiscreteDelta contains a QASimplescript with upper
   index position, fs represents the first script *)
upperQ[fs_QASimplescript] := MatchQ[fs, QASimplescript[_, _, True]]


(* Determines if first argument of a DiracDelta/DiscreteDelta contains a QASimplescript with upper
   index position *)
upperQ[fa_Plus] := MatchQ[fa, QASimplescript[_, _, True] + _] || MatchQ[fa, -QASimplescript[_, _, True] + _]


(* Determines if the arguments of a DiracDelta/DiscreteDelta match the required conditions for an
   invokation. *)
ruleMatchesQ[lst: { fa_, la___ }, a_Symbol, b_Symbol] := 
	Module[{pm},
		pm[x_] := x | -x; (* plus or minus an expression *)
		And @@ Map[MatchQ[lst[[#]], pm[QASimplescript[a, #, upperQ[fa]]] + pm[QASimplescript[b, #, upperQ[fa]]]]&, Range[Length[lst]]]
	]


(* Determines if the arguments of a DiracDelta/DiscreteDelta match the required conditions for an
   invokation. *)
ruleMatchesQ[lst: { fa_, la___ }, a_Symbol] := 
	Module[{pm},
		pm[x_] := x | -x; (* plus or minus an expression *)
		And @@ Map[MatchQ[lst[[#]], pm[QASimplescript[a, #, upperQ[fa]]] + _]&, Range[Length[lst]]]
	]


(* Determines the replacement rules for ommitting the DiracDelta/DiscreteDelta expression *)
getReplacement[lst: { fa_, la___ }, a_Symbol, b_Symbol] :=
	Module[{rules},
		rules = Map[Solve[lst[[#]] == 0, QASimplescript[b, #, upperQ[fa]]]&, Range[Length[lst]]] // Flatten;
		Append[rules, QASimplescript[b, 0, upperQ[fa]] -> QASimplescript[a, 0, upperQ[fa]]]
	]


(* Determines the replacement rules for ommitting the DiracDelta/DiscreteDelta expression
   TODO: role of time like variable?
*)
getReplacement[lst: { fa_, la___ }, a_Symbol] :=
	Module[{rules},
		rules = Map[Solve[lst[[#]] == 0, QASimplescript[a, #, upperQ[fa]]]&, Range[Length[lst]]] // Flatten;
		Append[rules, QASimplescript[a, 0, upperQ[fa]] -> 0]
	]


(* Determines if an expression is Skew Symmetric regarding an indexed variable *)
skewSymmetricQ[ex_, k_Symbol, dim_Integer, upper_] :=
	((ex + (ex /. 
		QASimplescript[k, i1_, upper] /; 
		(1 <= i1 <= dim) -> 
		(-QASimplescript[k, i1, upper]))) // QASimplify[#, Null, 3]&) == 0


(* Used to bring DiracDelta or DiscreteDelta variables in defined order *)
defineSortVariablesRule[continuous_, dim_Integer: 3] :=
	Module[{head, ex, h, k, up, part1, part2},
		head = If[continuous, DiracDelta, DiscreteDelta];
		ex = head @@ Table[QASimplescript[h, i, up] - QASimplescript[k, i, up], {i, 1, dim}];
		
		part1 = ex /. Map[# -> Pattern[#, Blank[]]&, {h, k, up}];
		part2 = ex /. (QASimplescript[h, i_, up] - QASimplescript[k, i_, up]) -> (QASimplescript[k, i, up] - QASimplescript[h, i, up]);
		
		Apply[HoldPattern[#1] :> #2 /; Order[#3, #4] < 0&, {part1, part2, h, k}]
	]


(* Determines if set (list) of indexed variables is negative *)
indicesNegative[lst_List, idx_Symbol, upper_, dim_Integer] := MatchQ[lst, Table[-QASimplescript[idx, i, upper], {i, 1, dim}]];


(* Determines if the empty or one-item sequence for adjunct is valid *)
adjunctValidQ [seq___] := ({ seq } == {} || seq == "H"); 


(* Replaces indexed variables with their negative counterpart *)
replaceNegative[k_Symbol, upper_, dim_Integer] := (QASimplescript[k, i_, upper] /; 1 <= i <= dim -> -QASimplescript[k, i, upper]);


(* Contracts an Integral expression *)
join[ex_, continuous_ : True] :=
	Module[{rule, ass, f, i, oldex, newex, pat, patph, path, patc, v, n, x, subex, integph, c, p, head, count},
		ass = <| |>;
		If[continuous,
			(head = QAIntegrate;
			pat = HoldPattern[QAIntegrate[integph, v: (_Symbol | _DirectedInfinity), n_Integer, x__Symbol]]),
			(head = QASum;
			pat = HoldPattern[QASum[integph, x__Symbol]])
		];
		path = HoldPattern[subex: Plus[Times[___ : 1, patph]..]] /. patph -> (pat /. integph -> Blank[]);
		patc = HoldPattern[Times[c___ : 1, patph]..] /. patph -> (pat /. integph -> Pattern[p, Blank[]]);
		
		rule = path :> 
			Module[{cases, exponent, func},
				QAPrint[StringTemplate["In replacement rule of `1`"][head], Indent->3];
				cases = Cases[subex, patc -> { c, p }, 1];
				exponent = Cases[subex, Power[E, expo_] -> expo, Infinity, 1];
				If[continuous,
					func = QAIntegrate[#, v, n, x]&,
					func = QASum[#, x]&
				];
				If[MatchQ[subex, _head] || MatchQ[subex, _Times],

					(QAPrint[StringTemplate["Simple `1` detected -> no replacement."][head], Indent->3];
					subex),

					(QAPrint[StringTemplate["Complex `1` -> perform replacement."][head], Indent->3];
					ass[f[exponent, i]] = func[Plus @@ Map[Times[Sequence @@ #]&, cases]];
					f[exponent, i])
				]
			];

		QAPrint[StringTemplate["Join `1` entered."][head]];
		oldex = ex;
		i = 1;
		While[i <= 10,
			If[!SyntaxQ[ToString[oldex, InputForm]],
				QAPrint["Not a valid expression.", Indent->2];
			];
			
			QAPrint[oldex, PrtLevel->3];
			QAPrint[StringTemplate["Join `2`, `1`. cycle"][i, head], Indent->2];
			count = Count[oldex, _head, Infinity];
			newex = If[count < 2, oldex, (QAPrint["Just before ReplaceAll, Count was ", count, Indent->2]; oldex /. { rule })];
			QAPrint["After If[count < 2]", Indent->2];
			
			If[Order[newex, oldex] == 0, Break[]];
			
			QAPrint["After If[Order == 0]", Indent->2];
			oldex = newex;
			i = i + 1;
		];
		
		QAPrint[StringTemplate["Join `1` leaving."][head]];
		newex /. f[lst_List, ii_Integer] :> ass[f[lst, ii]]
	]
 

End[]

EndPackage[]
