(* Mathematica package *)
(* Created by the Wolfram Workbench 14.04.2016 *)


BeginPackage[
	"QuantumAlgebra`Styles`",
	{
		"QuantumAlgebra`",
		"Notation`"
	}
]


(* Exported symbols added here with SymbolName::usage *) 

SetNotebookStyles::usage = 
	"SetNotebookStyles[] defines Notebook styles for the InputNotebook. \n" <>
	"SetNotebookStyles[nb] defines Notebook styles for the specified notebook. \n" <>
	"Calling this function once is required for proper display of the input / output " <> 
	"elements in Dirac Bra-Ket Notation.\n" <>
	"The original stylesheet is modified by adding some styles, but not changed completely."

SetDocumentationStyles::usage = 
	"SetDocumentationStyles[options] sets the documentation styles suitable for the " <>
	"QuantumAlgebra application's documentation so that Quantum objects can be displayed " <>
	"properly. This is usually not intended for the user."

App::usage = 
	"App option is used to overwrite the default."

Doc::usage = 
	"Doc option is used to overwrite the default."

FileSpec::usage = 
	"FileSpec option is used to overwrite the default."

Begin["`Private`"]
(* Implementation of the package *)


(* Sets the notebook styles for proper Quantum Algebra display *)
SetNotebookStyles[]:=
	Module[{nb},
		nb := InputNotebook[];
		SetNotebookStyles[nb]
	]

(* Set Notebook Styles
   NEW: 2020-04-23 
*)
SetNotebookStyles[doc_NotebookObject] :=
	
	Module[{styleDefinitions = CurrentValue[doc, StyleDefinitions], blankQ, base, cells, nb},
		blankQ[] := ! MatchQ[styleDefinitions, _Notebook];
		base[] := If[blankQ[], styleDefinitions, ""];
		
		cells = Which[
			blankQ[],																	(* fresh document: default / Documentation / other sheet *)
			QuantumAlgebra`Options`QAPrint["Blank Document"]; 
			{																			(* build cells list as needed by Quantum Algebra *)
				Cell[StyleData[StyleDefinitions -> base[]]],
				Sequence @@ styleCells[]
			},
			
			! quantumAlgebraStylesQ[styleDefinitions[[1]]],								(* we have already private styles, but not Quantum Algebra *)
			QuantumAlgebra`Options`QAPrint["Document with Styles, not Quantum"]; 
			Join[
				{ First[styleDefinitions[[1]]] }, 
				styleCells[], 															(* It is essential to insert the Quantum Algebra style cells *)
				Rest[styleDefinitions[[1]]]
			], 																			(* join existing cells with the Quantum Algebra style cells *)
			
			True,																		(* finally we have the Quantum Algebra style cells in document *)
			QuantumAlgebra`Options`QAPrint["Document with Styles, Quantum"]; 
			styleDefinitions[[1]]														(* extract the cells *)
		];

		nb = If[blankQ[],																(* build the stylesheet notebook *)
			Notebook[cells, StyleDefinitions -> "PrivateStylesheetFormatting.nb"],		(* modify or define options *)
			Notebook[cells, modifyOptions[styleDefinitions]]
		];
		SetOptions[doc, StyleDefinitions->nb]											(* and finally set it on document *)
	]


(* Options for SetDocumentationStyles *)
Options[SetDocumentationStyles]=
{
	App -> { $UserDocumentsDirectory, "Programmieren", "Mathematica", "Applications" },
	Doc -> { "QuantumAlgebra", "build" },
	FileSpec -> "*.nb"
}

(* SetDocumentationStyles[options] sets the documentation styles suitable for the QuantumAlgebra application's documentation *)
SetDocumentationStyles[OptionsPattern[]] :=
	Module[{setStyles, app, doc, dir, fileSpec, nbs},
	
		setStyles[nb_String] :=
			Module[{nbo},
				nbo = NotebookOpen[nb, Saveable->True];
				SetNotebookStyles[nbo];
				SetOptions[nbo, Saveable->False];
				NotebookSave[nbo];
				NotebookClose[nbo];
				Print /@ StringCases[nb, ___ ~~ "\\" ~~ fn: __ ~~ ".nb" :> (fn <> ".nb")];
		];
		
		app = OptionValue[App];
		doc = OptionValue[Doc];
		fileSpec = OptionValue[FileSpec];
		dir = FileNameJoin[Flatten[{app, doc}]];
		nbs = FileNames[fileSpec, {dir}, 10];
		Map[setStyles, nbs];
		True
	]


(* Modifies given Notebook's options *)
modifyOptions[nb_Notebook] :=
	Module[{opts, formattingnb},
		formattingnb = If[Count[nb, "Reference.nb", 10] > 0, "Default.nb", "PrivateStylesheetFormatting.nb"];
		opts = Options[nb];
		opts = opts /. (StyleDefinitions -> _) -> (StyleDefinitions -> formattingnb);
		opts
	]

(* Predicate, whether Quatum Algebra cells are present in list *)
quantumAlgebraStylesQ[cells: { _Cell ... }] := Count[cells, Cell[StyleData["BraKetArg"], ___], 10] > 0

(* Quantum Algebra Style Cells *)
styleCells[] := {
	
	Cell["Bra Ket Styles","Section"],
	Cell["\< The cells below define certain styles needed for defining Bras&Kets.These styles serve to give the correct structural properties to Bras & Kets \>","Text"],
	Cell[StyleData["BraKetArg"],
		AutoStyleOptions->{"UnmatchedBracketStyle"->"UnmatchedBracket"},
		SpanMinSize->Automatic,
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{0,0},{0,0}}},
		TagBoxOptions->{Editable->True,Selectable->True}],
	Cell[StyleData["BraWrapper"],
		AutoStyleOptions->{"UnmatchedBracketStyle"->None},
		SpanMinSize->2,
		ScriptBaselineShifts->{0.6,0.9},
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{0.2,-0.13},{0,0}}},
		TagBoxOptions->{Editable->False}],
	Cell[StyleData["KetWrapper"],
		AutoStyleOptions->{"UnmatchedBracketStyle"->None},
		SpanMinSize->2,
		ScriptBaselineShifts->{0.6,0.9},
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{-0.1,0.2},{0,0}}},
		TagBoxOptions->{Editable->False}],
	Cell[StyleData["BraKetWrapper"],
		AutoStyleOptions->{"UnmatchedBracketStyle"->None},
		SpanMinSize->2,
		ScriptBaselineShifts->{0.6,0.9},
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{0.25,0.25},{0,0}}},
		TagBoxOptions->{Editable->False}],
	Cell[StyleData["RIndex"],
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{-0.16,0},{0,0}}},
		TagBoxOptions->{Editable->False}],
	Cell[StyleData["LIndex"],
		StyleMenuListing->None,
		AdjustmentBoxOptions->{BoxMargins->{{0,-0.1},{0,0}}},
		TagBoxOptions->{Editable->False}]
}


(*
Alternative Options
{
	App \[Rule] {$UserBaseDirectory, "Applications"},
	Doc \[Rule] {"QuantumAlgebra", "Documentation"},
}
*)

End[]

EndPackage[]
